package Data::Squirrel;

use strict;
use vars qw($VERSION @ISA $VERBOSE);

$VERSION = '0.02';

use Apache::Session;
use Carp;
@ISA=qw(Apache::Session);

sub DEBUG {0}
$VERBOSE=0;

###########################################################
sub TIEHASH 
{
    my($package, $args) = @_;
    $args ||= {};

    # Make sure that the arguments to tie make sense
    
    if(ref $args ne "HASH") {
        die "Additional arguments should be in the form of a hash reference";
    }

    #Set-up the data structure and make it an object
    #of our class
    my $self = {
        args         => $args,
        lock         => 0,
        status       => 0,
        lock_manager => undef,  # These two are object refs ...
        object_store => undef,
        generate     => undef,  # but these three are subroutine refs
        serialize    => undef,
        unserialize  => undef,
    };
    
    bless $self, $package;
    $self->populate();
    return $self;
}

###########################################################
sub FETCH 
{
    my($self, $key) = @_;
    local $self->{data}={};
    local $self->{serialized};
    $self->{data}{_session_id}=$key;        
    return unless $self->{object_store}->exists($self);

    $self->acquire_read_lock;
    eval {$self->{object_store}->materialize($self)};
    DEBUG and $@ and warn $@;
    $self->release_read_lock;
    $self->{unserialize}->($self);
    return $self->{data}{DATA};
}

###########################################################
sub STORE 
{
    my($self, $key, $data) = @_;
    local $self->{data}={DATA=>$data};
    local $self->{serialized};
    $self->{data}{_session_id}=$key;                    
    $self->{data}{TIME}=time;

    $self->{serialize}->($self);

    $self->acquire_write_lock;
    eval {
        if($self->{object_store}->exists($self)) {
            $self->{object_store}->update($self);
        } else {
            $self->{object_store}->insert($self);
        }
    };
    $self->release_write_lock;
    die $@ if $@;

    return $self->{data}{DATA};
}

###########################################################
sub DELETE 
{
    my($self, $key)=@_;
    local $self->{data}={};
    $self->{data}{_session_id}=$key;
    return unless $self->{object_store}->exists($self);

    local $self->{serialized};
    $self->FETCH($key);     # get object into {data}{DATA}

    $self->acquire_write_lock;
    eval { 
        $self->{object_store}->remove($self);
    };
    $self->release_write_lock;
    die $@ if $@;

    return $self->{data}{DATA};

}

###########################################################
sub EXISTS 
{
    my($self, $key)=@_;
    local $self->{data}={};
    DEBUG and warn "Looking for $key\n";
    $self->{data}{_session_id}=$key;        
    $self->acquire_read_lock;
    my $ret=$self->{object_store}->exists($self);
    $self->release_read_lock;
    DEBUG and warn $ret ? "Found\n" : "Not found\n";
    return $ret;
}

###########################################################
sub FINDFIRST
{
    my($self)=@_;
    croak ref($self), "FINDFIRST not implemented";
}

###########################################################
sub FINDNEXT
{
    my($self)=@_;
    croak ref($self), "::FINDNEXT not implemented";
}

###########################################################
sub DESTROY 
{
    my($self) = @_;
    
    $self->release_all_locks;
}

sub expire 
{
    my($self, $expire)=@_;

    unless(defined $expire) {
        warn "Must specify expiry time";
        return;
    }


    #####################################################
    # clean up the object store
    my $ids=$self->list_ids();
    my $now = time;
    DEBUG and warn "It is now $now\n";
    foreach my $id (@$ids) { 

        local $self->{data};
        $self->{data}{_session_id}=$id;

        DEBUG and warn "Checking $id\n";

        $self->acquire_read_lock;
        eval {
            local $self->{serialized};                  # get storage date
            $self->{object_store}->materialize($self);  
            $self->{unserialize}->($self);
        };
        if($@) {
            warn $@;
            $self->release_read_lock;
            next;
        }
        next unless $self->{data} and defined $self->{data}{TIME};                            
        DEBUG and warn "Created at $self->{data}{TIME}\n";
        next unless ($now - $self->{data}{TIME}) > $expire;
        DEBUG and warn "Too old ($expire)\n";

        $self->acquire_write_lock;                      # delete if too old
        eval { 
            $VERBOSE and warn "Expiring $id";
            $self->{object_store}->remove($self);
        };
        warn $@ if $@;
        $self->release_write_lock;
        $self->release_read_lock;
    }

    #####################################################
    # clean up the lock manager
    if($self->{lock_manager}->can('clean')) {
        my $directory = $self->{args}->{LockDirectory} || $Apache::Session::Lock::File::Directory;
        die "The directory for object storage must be passed as an argument or set in httpd.conf" 
                unless defined $directory;
        $self->{lock_manager}->clean($directory, $expire);
    } 
    elsif($self->{lock_manager}->can('expire')) {
        $self->{lock_manager}->expire($expire);
    }

}



1;
__END__
# Below is the stub of documentation for your module. You better edit it!

=head1 NAME

Data::Squirrel - Store arbitrary data anywhere Apache::Session can

=head1 SYNOPSIS

  use Data::Squirrel;
  tie %squirrel, "Data::Squirrel::MySQL", {DataSource=>"dbi:mysql:test", 
          UserName=>"test", Password=>"test", 
            LockDataSource=>"dbi:mysql:test", 
            LockUserName=>"test", LockPassword=>"test", };

  $squirrel{$id}=$somedata;



=head1 DESCRIPTION

Data::Squirrel is a wrapper around the Apache::Session modules and borrows
heavily it's design.  It allows you to store data using a consistent
interface (tied hash).  You don't use Data::Squirrel directly, but rather
one of the sub-modules, so go look at 'em!  It also adds object expiration.

=head1 INTERFACE

=head2 tieing a section

Create a store using files:

    tie %squirrel, 'Data::Squirrel::MySQL, 
                        {Directory=>'/var/something/mydata', 
                         LockDirectory=>'/tmp'},


Get one block of data:

    my $data=$squirrel{$id};

Put some data in:

    $squirrel{$id}=$data;

Woah, Nelly!

    delete $squirrel{$id}       # data is NO MORE

Is there anybody out there?

    if(exists $squirrel{$id}) {
        # yo, we gotcha
    }

Using exists() is faster then a fetch, because you don't have to unserialise
the data.

Do you like lists?

    print "This is all I have:\n";
    foreach my $id (keys %squirrel) {
        print "$id\n";
    }

=head1 METHODS

=head2 expire($seconds)

Deletes all objects in the store that were written to over $seconds seconds
ago.  Say you use Data::Squirrel to hold session info for a web application. 
You don't really know when the user walks away and won't come back.  So you
expire unused session info every 3 hours (say).

    tied($sessions)->expire(3 * 60 * 60);

This operation grabs and releases the read and write locks continually, so
don't call it to often.

=head1 OVERLOADING

The Squirrel modules use Apache::Session::Generate::MD5 and
Apache::Session::Serialise::Storable by default.  If you want to use
something else, just overload C<populate>:

    package Data::Squirrel::Mine;
    use strict;
    
    use Data::Denter;
    use vars qw(@ISA);
    @ISA=qw(Data::Squirrel::File);

    sub populate
    {
        my($self)=@_;
        $self->SUPER::populate;
        $self->{generate}     = \&gen_key;
        $self->{validate}     = \&validate_key;
        $self->{serialize}    = \&dent;
        $self->{unserialize}  = \&undent;   
    }

    sub gen_key
    {
        # i don't think this will ever be called.
    }

    sub validate_key
    {
        my($self)=@_;
        if($self->{data}{_session_id} =~ m(^([0-9a-f]+)$)i) {
            $self->{data}{_session_id}=$1;  # untainting too!
            return;
        }
        die "NOT VALID";            # yuck!  calls are wraped in eval {}
    }

    sub dent
    {
        my($self)=@_;
        $self->{serialized}=Denter $self->{data};
    }

    sub undent
    {
        my($self)=@_;
        $self->{data}=Denter $self->{serialized};
    }

Yes, this is a strange way to propagate parameters.  It's not my idea.

You can also overload  {object_store} and {lock_manager}.  See the source
for the interface.  If you overload {object_store}, be sure to also create a
list_ids() method.


=head1 BUGS

There is a BUG in Apache::Session that causes it to die when storing to a
key that doesn't exist.  This baffles me.

Also, fetching and storing data can be relatively expensive, so you should
do it as rarely as possible.  This means doing things like :

    my $something=$squirrel{$key};
    foreach my $n (@foo) {
        $something->{$n}++;
    }
    $squirrel{$key}=$something;


=head1 SEE ALSO

=head1 AUTHOR

Philip "Leolo" Gwyn, fil@pied.nu

=head1 SEE ALSO

Apache::Session, perl(1),  L<Data::Squirrel::MySQL>,
L<Data::Squirrel::File>, L<Data::Squirrel::DB_File>,
L<Data::Squirrel::GDBM_File>.

=cut
